/*  Filename: TabLang.js
 * $Revision: 1.2 $
 *     $Date: 2008/08/05 20:15:00 $
 *
 * TabLang-related code for TabLang service (global XPCOM object)
 * This file is loaded as subscript from components/TabLangService.js
 * Implementations of TabLang (main extension object), Options and
 * TabLangSwitcher wrapper objects go here.
 * 
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TabLang extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is Andrey Lokot.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrey Lokot <crazy-eye@yandex.ru>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// To access main TabLang object use
// Components.classes["tablang@mozdev.org/tablangservice;1"].getService().wrappedJSObject.tablang

// alert() is a shortcut to show message to the user ( replacement of window.alert() )
// Parameters:
//       str - string with message to show
//   caption - caption of message window (optional, "TabLang" is default)
//       win - window message will popup from (optional, current window is default)
function alert(str, caption, win) {
	Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
		.getService(Components.interfaces.nsIPromptService)
		.alert(win ? win : null, caption ? caption : "TabLang", str);
}

// reportError() produces error message tin Error Console and shows it to the user
// Parameters:
//       str - string with error message
//       win - window message for the user will popup from (optional, current window is default).
//             Set to "false" for silent error report (error will be shown in console only)
function reportError(str, win) {
	Components.utils.reportError("[TabLang] " + str);
	if (win !== false) {
		alert(str, "TabLang error", win);
	}
}

// Wrapper for TabLangSwitcher component. Object operates with variables of (informal) type LANG
// which can be both LANG_TAG and LANG_ID so it makes translation between LANG_TAG and LANG_ID
// transparent for the caller. "switcher" member can be used to access TabLangSwitcher component directly.
function SwitcherWrapper() {
	// "switcher" variable is used to access TabLangSwitcher component (through ITabLangSwitcher interface)
	this.switcher = Components.classes["tablang@mozdev.org/tablangswitcher;1"]
						.getService(Components.interfaces.ITabLangSwitcher);
}
SwitcherWrapper.prototype = {
	// "currentLang" variable of type LANG which is used to get and set current system input language
	get currentLang () {
		// "currentLang" getter returns current system LANG. It uses
		// GetCurrentLangId() so actual return value is LANG_ID.
		try {
			return this.switcher.GetCurrentLangId();
		}
		catch (err) {
			reportError("Exception in currentLang getter\n" + err);
			// Trying to hide failure
			return Components.classes["tablang@mozdev.org/tablangservice;1"]
						.getService().wrappedJSObject.options
						.defaultOverride;
		}
	},
	set currentLang (lang) {
		// "currentLang" setter takes "lang" parameter of type LANG. Uses SetCurrentLangId() 
		// if "lang" is LANG_ID or SetCurrentLangTag() if "lang" is LANG_TAG
		if (lang == undefined) {
			// Possibly due to bug 366705 (https://bugzilla.mozilla.org/show_bug.cgi?id=366705):
			// TabOpen may not fire when user tries to load nonexisting page from the local drive
			// so browser.tablangLang would not be set for this tab. This was fixed in Firefox 3.0b1
			reportError(
				"Language not defined for current tab",
				// Alerting only if in Firefox 3.0b1 or later
				( Components.classes["@mozilla.org/xpcom/version-comparator;1"]
						.getService(Components.interfaces.nsIVersionComparator).
						compare(
							Components.classes["@mozilla.org/xre/app-info;1"]
								.getService(Components.interfaces.nsIXULAppInfo)
								.version,
							"3.0b1"
					) >= 0 ) ? undefined : false
			);
			lang = Components.classes["tablang@mozdev.org/tablangservice;1"]
						.getService().wrappedJSObject.options
						.defaultOverride;
		}
		try {
			if (typeof(lang) == "number") {
				this.switcher.SetCurrentLangId(lang);
			}
			else {
				this.switcher.SetCurrentLangTag(lang);
			}
		}
		catch (err) {
			reportError("Exception in currentLang setter\nnewlang=" + lang + "\n" + err);
			// New language wasn't set but we still can try to continue execution
		}
	},
	// forceLangTag() function converts LANG to LANG_TAG
	// Takes "lang" parameter of type LANG and returns corresponding LANG_TAG
	// If "lang" is actually LANG_TAG returns "lang" itself. Otherwise
	// uses LangTagFromId() to convert LANG_ID to LANG_TAG.
	forceLangTag: function(lang) {
		try {
			return (typeof(lang) == "number") ? this.switcher.LangTagFromId(lang) : lang;
		}
		catch (err) {
			reportError("Exception in forceLangTag()\nlang=" + lang + "\n" + err);
			// Seems to be the error in caller code so throwing exception out
			throw err;
		}
	}
}

// Options object is used to interact with extension properties
function Options() {
	// *** initialization helper functions ***
	// getLangList() wraps TabLangSwitcher.ListLangTags() and
	// returns list of LANG_TAGs (array of strings)
	function getLangList() {
		var list = { value: [] };
		var count = { value: list.value.count };
		Components.classes["tablang@mozdev.org/tablangswitcher;1"]
			.getService(Components.interfaces.ITabLangSwitcher)
			.ListLangTags(count, list);
		return list.value;
	}
	// *** properties ***
	// prefs - preferences service instance for extension(!) properties
	this.prefs = Components.classes["@mozilla.org/preferences-service;1"]
					.getService(Components.interfaces.nsIPrefService)
					.getBranch("extensions.tablang.");
	// langList - list of languages available in the running system
	this.langList = getLangList();
	// defaultOverride - language used instead of one in preferences if that is not available
	this.defaultOverride = null;
	// If defaultLang is in the list validated() will return defaultLang
	// Otherwise it will return defaultOverride == null, we will change it 
	// to langList[0] (there should be at least one language in the system)
	this.defaultOverride = this.defaultLang;
	if (this.defaultOverride == null) {
		this.defaultOverride = this.langList[0];
	}
}
Options.prototype = {
	// validated() - replaces the language if it is not availabe with defaultOverride
	// Takes "lang" - LANG_TAG to validate.
	// Returns "lang" if it is availabe or defaultOverride otherwise
	validated: function(lang) {
		return (this.langList.indexOf(lang) != -1) ? lang : this.defaultOverride;
	},
	// *** The following properties are correct(!) LNAG_TAGs of corresponding preferences ***
	// defaultLang - the language to be used for new tabs (with no parent)
	get defaultLang () { return this.validated(this.prefs.getCharPref("defaultLang")); },
	// urlbarLang - the language to be used when editing URLs in address bar
	get urlbarLang () { return this.validated(this.prefs.getCharPref("urlbarLang")); }
};

// TabLang is the main extension object. It contains all the
// functions to work with "objects" which language is controlled
function TabLang() {
	// TabLangSwitcher wrapper instance
	this.wrapper = new SwitcherWrapper;
	// Options instance
	this.options = new Options;
	// currentObject - currently focused object (at least one extension supports).
	// Current system input language is implied to be the language of this object
	// Don't change it manually. To switch to another object use switchTo()
	this.currentObject = null;
	// lastBrowser - last active (focused) browser: the last object
	// of "browser" type that was (or is) saved in currentObject
	// May be not focused. Is used to find "parent" for new tab
	// Don't change it manually. To switch to another object use switchTo()
	this.lastBrowser = null;
	// urlbarLang - the language of urlbars of all windows. 
	this.urlbarSharedLang = {};
	this.attachLang(this.urlbarSharedLang, this.options.urlbarLang);
}
TabLang.prototype = {
	// attachLang() - adds tablangLang property to the object.
	// It should be called before switching to this object.
	// Parameters:
	//    obj - object to attach tablangLang property to
	//   lang - initial (stored) language of the object
	attachLang: function (obj, lang) {
		var tablang = this;
		// tablangInternalLang - stored language. It is used when oject is not focused
		obj.tablangInternalLang = lang;
		// tablangLang - property, representing language of the object.
		// If object is focused gets and sets current system language
		// else gets and sets tablangInternalLang
		obj.__defineGetter__("tablangLang", function() {
			return (tablang.currentObject == this) ? 
				tablang.wrapper.currentLang : this.tablangInternalLang;
		});
		obj.__defineSetter__("tablangLang", function(newlang) {
			if (tablang.currentObject == this) {
				tablang.wrapper.currentLang = newlang;
			}
			// Always saving to tablangInternalLang
			// (in case something will go wrong)
			this.tablangInternalLang = newlang;
		});
	},
	// switchTo() - switches internal extension focus to another object.
	// Use it instead of manually changing currentObject/lastBrowser
	// Parameters:
	//      newobj - object to switch internal focus to
	//   isBrowser - should be true/false depending on if newobj is browser.
	//               lastBrowser value is updated accordingly.
	switchTo: function (newobj, isBrowser) {
		if (this.currentObject) {
			// If there was previous focused object saving its language
			this.currentObject.tablangInternalLang = this.wrapper.currentLang;
		}
		// Switching internally to new object
		this.currentObject = newobj;
		if (isBrowser) {
			this.lastBrowser = newobj;
		}
		// Loading saved language for selected tab
		if (newobj) {
			this.wrapper.currentLang = newobj.tablangInternalLang;
		}
	},
	// setEventHandlers() sets event handlers for all
	// events in this window extension supports
	// Parameters:
	//   win - window to set event handlers for
	setEventHandlers: function(win) {
		var tablang = this;
		win.gBrowser.addEventListener(
			"focus",
			function(event) { tablang.onSelectTab(win); },
			true);
		win.gBrowser.tabContainer.addEventListener(
			"select",
			// Switching language also when switching between tabs when tabbar is focused
			function(event) {
				if (tablang.currentObject == tablang.lastBrowser) {
					tablang.onSelectTab(win);
				}
			},
			false);
		win.gBrowser.tabContainer.addEventListener(
			"TabOpen",
			function(event) {
				tablang.onNewTab(win, event.target.linkedBrowser);
			},
			false);
		win.document.getElementById("appcontent").addEventListener(
			"DOMContentLoaded",
			function(event) {
				var doc = event.originalTarget;
				// Calling onPageLoad() only for top documents (not frames/iframes)
				if (doc.nodeName == "#document" && doc.defaultView.top == doc.defaultView) {
					tablang.onPageLoad(win, win.gBrowser.getBrowserForDocument(doc));
				}
			},
			true);
		win.gBrowser.tabContainer.addEventListener(
			"TabClose",
			function(event) { tablang.onTabClose(win, event.target.linkedBrowser); },
			false);
		win.addEventListener(
			"close",
			function(event) { tablang.onWindowClose(win); },
			false);
		var urlbarInput = win.document.getElementById("urlbar").inputField;
		urlbarInput.addEventListener(
			"focus",
			function(event) { tablang.onUrlbarFocus(win); },
			false);
		urlbarInput.addEventListener(
			"blur",
			function(event) { tablang.onUrlbarBlur(win); },
			false);
	},
	// *** The following functions are event handlers. They take these input parameters:
	// ***       win - window where event occured
	// ***   browser - browser object for which the event was sent (for tab events)
	// onSelectTab() is called every time any element inside tabbrowser
	// receives focus as focus events for browser objects are not sent sometimes
	onSelectTab: function(win) {
		// Checking if focus was actually received by 
		// some tab but not the active one (if any)
		if (this.currentObject != win.gBrowser.selectedBrowser) {
			this.switchTo(win.gBrowser.selectedBrowser, true);
		}
	},
	// onNewTab() is called when new tab is created. 
	// Initializes new tab and attaches TabLangState to it
	onNewTab: function(win, browser) {
		this.attachLang(browser, this.options.defaultLang);
		if (this.lastBrowser) {
			browser.tablangSavedState = {
				/*browser: this.lastBrowser,*/
				uri: this.lastBrowser.currentURI, 
				lang: this.lastBrowser.tablangLang
			};
		}
		win.setTimeout(function() {
					// If window.opener is defined use it as parent window
					if (browser.contentWindow.opener) {
						var openerBrowser = null;
						var doc = browser.contentWindow.opener.top ?
							browser.contentWindow.opener.top.document : browser.contentWindow.opener.document;
						var winEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"]
											.getService(Components.interfaces.nsIWindowMediator)
											.getEnumerator("navigator:browser");
						while (!openerBrowser && winEnum.hasMoreElements()) {
							var enumWindow = winEnum.getNext();
							if (enumWindow.gBrowser) {
								openerBrowser = enumWindow.gBrowser.getBrowserForDocument(doc);
							}
						}
						if (openerBrowser) {
							browser.tablangLang = openerBrowser.tablangLang;
							delete browser.tablangSavedState;
						}
					}
			}, 0);
	},
	// onPageLoad() is called after DOM structure for the page is created
	// so we can try to use document.referrer to find tab from which
	// current page was opened (parent tab)
	onPageLoad: function(win, browser) {
		if (browser.tablangSavedState) {
			// done will be set to true when we find parent and no further search is needed
			var done = false;
			// Trying to find parent window using document.referrer
			if (browser.contentDocument.referrer) {
				var referrerURI = Components.classes["@mozilla.org/network/io-service;1"]
								.getService(Components.interfaces.nsIIOService)
								.newURI(browser.contentDocument.referrer, null, null);
				if (browser.tablangSavedState.uri.equals(referrerURI)) {
					browser.tablangLang = browser.tablangSavedState.lang;
					done = true;
				}
			}
			// Delete info saved when tab was opened as it is useless now
			delete browser.tablangSavedState;
		}
	},
	// onTabClose() is called when tab is closed to prevent
	// variables from pointing to incorrect browser object
	onTabClose: function(win, browser) {
		if (this.lastBrowser == browser) {
			if (this.currentObject == this.lastBrowser) {
				this.currentObject = null;
			}
			this.lastBrowser = null;
		}
	},
	// onWindowClose() is called when window is closed to prevent
	// variables from pointing to incorrect browser object
	onWindowClose: function(win) {
		if (this.lastBrowser && 
				win.gBrowser.getBrowserForDocument(this.lastBrowser.contentDocument)) {
			if (this.currentObject == this.lastBrowser) {
				this.currentObject = null;
			}
			this.lastBrowser = null;
		}
	},
	// onUrlbarFocus() is called when user begins using address
	// bar. It switches current language to urlbarLang
	onUrlbarFocus: function(win) {
		this.switchTo(this.urlbarSharedLang, false);
	},
	// onUrlbarBlur() is called when user leaves address bar. It saves
	// urlbarLang as it should not be affected by the language user
	// may use in another controls that are not supported
	// (we assume user mostly uses English in urlbar)
	onUrlbarBlur: function(win) {
		if (this.currentObject == this.urlbarSharedLang) {
			this.switchTo(null, false);
		}
	}
};

// initTabLangService() is called from components/TabLangService.js
// It should populate service object "obj" with TabLang-related members
function initTabLangService(obj) {
	// tablang - global TabLang instance
	obj.tablang = new TabLang;
	// switcherWrapper - shortcut to tablang.wrapper
	obj.switcherWrapper = obj.tablang.wrapper;
	// options - shortcut to tablang.options
	obj.options = obj.tablang.options;
	// *** DEBUG ***
	// debugReInit() restarts extension. Use for debug only!
	// All browser windows should be closed before calling it.
	// Is used in TabLang tests.
	obj.debugReInit = function() {
		this.switcherWrapper = null;
		this.options = null;
		this.tablang = new TabLang;
		this.switcherWrapper = this.tablang.wrapper;
		this.options = this.tablang.options;
	};
}
